! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2011 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!                     Meta data subdomain                      !
!                                                              !
!--------------------------------------------------------------!

#include "fd4.h"

#ifndef FD4_COUPLE_MAX_PARTITIONS
#define FD4_COUPLE_MAX_PARTITIONS 2
#endif

!! A subdomain of meta data for coupling.
!!
!! @author Matthias Lieber
module fd4_mdsd_mod

  use fd4_globaldef_mod
  use fd4_mpi_mod
  use stack3_mod
  
  USE_MPI
  
  implicit none
  private
  
  INCLUDE_MPIFH

  !! max. number of coupled partitions per meta data subdomain
  integer, parameter, public :: FD4_MD_MAX_PART = FD4_COUPLE_MAX_PARTITIONS
  
  !! flags used when serializing a mdsd
  !integer, parameter :: MDSD_DESC_RANK=1, MDSD_DESC_VAR=2, MDSD_DESC_ABND=4
  

  public :: fd4_mdsd, fd4_mdsd_free
  !public :: fd4_mdsd_serialize, fd4_mdsd_unserialize


  !! Type for a single meta data subdomain.
  !! 
  !! A meta data subdomain is a rectangular subsection of the block decomposition
  !! and contains all metadata that is located or begins within this subdomain.
  type fd4_mdsd
    type(stack3) :: desc_rank(FD4_MD_MAX_PART)      !! rank, start and end index in $desc_var (one stack3 entry)
    type(stack3) :: desc_abnd(FD4_MD_MAX_PART)      !! abnd (two stack3 entries) of a couple array
    !logical :: serialized = .false.                 !! .true. if all stacks of the mdsd are serialized together in one array
  end type fd4_mdsd


contains

  include 'util/stack3.incf'

  subroutine fd4_mdsd_free(mdsd)
    type(fd4_mdsd), pointer :: mdsd
    integer(i_k) :: idx
    do idx=1,FD4_MD_MAX_PART
      call stack3_delete(mdsd%desc_rank(idx))
      call stack3_delete(mdsd%desc_abnd(idx))
    end do
    deallocate(mdsd)
    NULLIFY(mdsd)
  end subroutine fd4_mdsd_free

#ifdef COMMENT
  !! Serialize all $stack3s of the mdsd so that they together use one single array.
  !!
  !! *CAUTION*: Do not call any of the stack3s subroutines for the stacks of a serialized mdsd.
  subroutine fd4_mdsd_serialize(mdsd)
    type(fd4_mdsd), pointer :: mdsd
    integer(i_k) :: idx, i, len
    integer(i_k), pointer :: m(:,:)

    if(mdsd%serialized) return
    
    ! determine length of the array and allocate array
    len = mdsd%desc_abnd%pos + 1
    do idx=1,FD4_MD_MAX_PART
      if(associated(mdsd%desc_rank(idx)%pos)) then
        len = len + mdsd%desc_rank(idx)%pos + mdsd%desc_var(idx)%pos + 2
      end if
    end do
    allocate(m(3,len))
    ! copy contents from the original stack arrays to the new array,
    ! deallocate the original data array of the stacks,
    ! and bend the pointers of the original stacks to point into the new created array
    i=1
    ! desc_abnd
    m(1,i) = mdsd%desc_abnd%pos
    m(2,i) = MDSD_DESC_ABND
    m(1:3,i+1:i+mdsd%desc_abnd%pos) = mdsd%desc_abnd%a(1:3,1:mdsd%desc_abnd%pos)
    mdsd%desc_abnd%a   => m(1:3,i+1:i+mdsd%desc_abnd%pos)
    mdsd%desc_abnd%pos => m(1,i)
    deallocate(mdsd%desc_abnd%m)
    mdsd%desc_abnd%m   => m
    i=i+mdsd%desc_abnd%pos+1
    do idx=1,FD4_MD_MAX_PART
      if(associated(mdsd%desc_rank(idx)%pos)) then
        ! desc_rank
        m(1,i) = mdsd%desc_rank(idx)%pos
        m(2,i) = MDSD_DESC_RANK
        m(3,i) = idx
        m(1:3,i+1:i+mdsd%desc_rank(idx)%pos) = mdsd%desc_rank(idx)%a(1:3,1:mdsd%desc_rank(idx)%pos)
        mdsd%desc_rank(idx)%a   => m(1:3,i+1:i+mdsd%desc_rank(idx)%pos)
        mdsd%desc_rank(idx)%pos => m(1,i)
        deallocate(mdsd%desc_rank(idx)%m)
        mdsd%desc_rank(idx)%m   => m(1:3,i  :i+mdsd%desc_rank(idx)%pos)
        i=i+mdsd%desc_rank(idx)%pos+1
        ! desc_var
        m(1,i) = mdsd%desc_var(idx)%pos
        m(2,i) = MDSD_DESC_VAR
        m(3,i) = idx
        m(1:3,i+1:i+mdsd%desc_var(idx)%pos) = mdsd%desc_var(idx)%a(1:3,1:mdsd%desc_var(idx)%pos)
        mdsd%desc_var(idx)%a   => m(1:3,i+1:i+mdsd%desc_var(idx)%pos)
        mdsd%desc_var(idx)%pos => m(1,i)
        deallocate(mdsd%desc_var(idx)%m)
        mdsd%desc_var(idx)%m   => m(1:3,i  :i+mdsd%desc_var(idx)%pos)
        i=i+mdsd%desc_var(idx)%pos+1
      end if
    end do
    FD4_L3_ASSERT(len/=i-1,'inconsistency in len '//i2s(len)//i2s(i))
    mdsd%serialized = .true.

  end subroutine fd4_mdsd_serialize


  !! Unserialize all $stack3s of the mdsd
  !!
  subroutine fd4_mdsd_unserialize(mdsd)
    type(fd4_mdsd), pointer :: mdsd
    integer(i_k) :: idx, i, len
    integer(i_k), pointer :: m(:,:)

    if(.not. mdsd%serialized) return
    
    ! get the pointer to the array containg all the data of the stack
    m => mdsd%desc_abnd%m
    len = ubound(m,2)
    ! desc_abnd
    i=1
    FD4_L3_ASSERT(m(2,i)/=MDSD_DESC_ABND,'inconsistency in stack array type ABND')
    allocate(mdsd%desc_abnd%m(3,0:mdsd%desc_abnd%pos))
    mdsd%desc_abnd%m(1:3,0:mdsd%desc_abnd%pos) = m(1:3,i:i+mdsd%desc_abnd%pos)
    if(mdsd%desc_abnd%pos>0) mdsd%desc_abnd%a => mdsd%desc_abnd%m(1:3,1:mdsd%desc_abnd%pos)
    mdsd%desc_abnd%pos => mdsd%desc_abnd%m(1,0)
    i=i+mdsd%desc_abnd%pos+1

    do idx=1,FD4_MD_MAX_PART
      if(i>len) exit
      if(m(3,i)<idx) cycle
      FD4_L3_ASSERT(m(3,i)>idx,'inconsistency in stack array idx')
      ! desc_rank
      FD4_L3_ASSERT(m(2,i)/=MDSD_DESC_RANK,'inconsistency in stack array type RANK')
      allocate(mdsd%desc_rank(idx)%m(3,0:mdsd%desc_rank(idx)%pos))
      mdsd%desc_rank(idx)%m(1:3,0:mdsd%desc_rank(idx)%pos) = m(1:3,i:i+mdsd%desc_rank(idx)%pos)
      if(mdsd%desc_rank(idx)%pos>0) mdsd%desc_rank(idx)%a => mdsd%desc_rank(idx)%m(1:3,1:mdsd%desc_rank(idx)%pos)
      mdsd%desc_rank(idx)%pos => mdsd%desc_rank(idx)%m(1,0)
      i=i+mdsd%desc_rank(idx)%pos+1
      ! desc_var
      FD4_L3_ASSERT(m(2,i)/=MDSD_DESC_VAR,'inconsistency in stack array type VAR')
      allocate(mdsd%desc_var(idx)%m(3,0:mdsd%desc_var(idx)%pos))
      mdsd%desc_var(idx)%m(1:3,0:mdsd%desc_var(idx)%pos) = m(1:3,i:i+mdsd%desc_var(idx)%pos)
      if(mdsd%desc_var(idx)%pos>0) mdsd%desc_var(idx)%a => mdsd%desc_var(idx)%m(1:3,1:mdsd%desc_var(idx)%pos)
      mdsd%desc_var(idx)%pos => mdsd%desc_var(idx)%m(1,0)
      i=i+mdsd%desc_var(idx)%pos+1
    end do
    deallocate(m)
    FD4_L3_ASSERT(len/=i-1,'inconsistency in len '//i2s(len)//i2s(i))
    mdsd%serialized = .false.

  end subroutine fd4_mdsd_unserialize

#endif

end module fd4_mdsd_mod
